@@ -5,6 +5,7 @@ mod cd;
 mod cp_mv;
 mod exit;
 mod mkdir;
+mod pwd;
 mod rm;
 mod sleep;
 
@@ -12,5 +13,6 @@ pub use cd::*;
 pub use cp_mv::*;
 pub use exit::*;
 pub use mkdir::*;
+pub use pwd::*;
 pub use rm::*;
 pub use sleep::*;
