@@ -17,9 +17,13 @@ use crate::commands::pwd_command;
 use crate::commands::rm_command;
 use crate::commands::sleep_command;
 use crate::parser::Command;
-use crate::parser::PipelineOperator;
+use crate::parser::PipeSequence;
+use crate::parser::PipeSequenceOperator;
+use crate::parser::Pipeline;
+use crate::parser::PipelineInner;
 use crate::parser::Sequence;
 use crate::parser::SequentialList;
+use crate::parser::SimpleCommand;
 use crate::parser::StringOrWord;
 use crate::parser::StringPart;
 use crate::shell_types::pipe;
@@ -161,9 +165,6 @@ fn execute_sequence(
         )],
         Vec::new(),
       ),
-      Sequence::Command(command) => {
-        execute_command(command, state, stdin, stdout, stderr).await
-      }
       Sequence::BooleanList(list) => {
         let mut changes = vec![];
         let first_result = execute_sequence(
@@ -219,91 +220,146 @@ fn execute_sequence(
         }
       }
       Sequence::Pipeline(pipeline) => {
-        let mut wait_tasks = vec![];
-        let mut last_output = Some(stdin);
-        let mut next_sequence = Some(Sequence::Pipeline(pipeline));
-        while let Some(sequence) = next_sequence.take() {
-          let (output_reader, output_writer) = pipe();
-          let (stderr, sequence) = match sequence {
-            Sequence::Pipeline(pipeline) => {
-              next_sequence = Some(pipeline.next);
-              (
-                match pipeline.op {
-                  PipelineOperator::Stdout => stderr.clone(),
-                  PipelineOperator::StdoutStderr => output_writer.clone(),
-                },
-                pipeline.current,
-              )
-            }
-            _ => (stderr.clone(), sequence),
-          };
-          wait_tasks.push(execute_sequence(
-            sequence,
-            state.clone(),
-            last_output.take().unwrap(),
-            output_writer.clone(),
-            stderr.clone(),
-          ));
-          last_output = Some(output_reader);
-        }
-        let output_handle = tokio::task::spawn_blocking(|| {
-          last_output.unwrap().pipe_to_sender(stdout).unwrap();
-        });
-        let mut results = futures::future::join_all(wait_tasks).await;
-        output_handle.await.unwrap();
-        let last_result = results.pop().unwrap();
-        let all_handles = results.into_iter().flat_map(|r| r.into_handles());
-        match last_result {
-          ExecuteResult::Exit(code, mut handles) => {
-            handles.extend(all_handles);
-            ExecuteResult::Continue(code, Vec::new(), handles)
-          }
-          ExecuteResult::Continue(code, _, mut handles) => {
-            handles.extend(all_handles);
-            ExecuteResult::Continue(code, Vec::new(), handles)
-          }
-        }
-      }
-      Sequence::Subshell(list) => {
-        let result = execute_sequential_list(
-          *list,
-          state.clone(),
-          stdin,
-          stdout,
-          stderr,
-          // yield async commands to the parent
-          AsyncCommandBehavior::Yield,
-        )
-        .await;
-
-        // sub shells do not cause an exit
-        match result {
-          ExecuteResult::Exit(code, handles) => {
-            ExecuteResult::Continue(code, Vec::new(), handles)
-          }
-          ExecuteResult::Continue(_, _, _) => result,
-        }
-      }
-      Sequence::Negated(sequence) => {
-        let result =
-          execute_sequence(*sequence, state, stdin, stdout, stderr).await;
-        match result {
-          ExecuteResult::Exit(code, handles) => {
-            ExecuteResult::Exit(code, handles)
-          }
-          ExecuteResult::Continue(code, changes, handles) => {
-            let new_code = if code == 0 { 1 } else { 0 };
-            ExecuteResult::Continue(new_code, changes, handles)
-          }
-        }
+        execute_pipeline(pipeline, state, stdin, stdout, stderr).await
       }
     }
   }
   .boxed()
 }
 
-async fn execute_command(
-  command: Command,
+async fn execute_pipeline(
+  pipeline: Pipeline,
+  state: ShellState,
+  stdin: ShellPipeReader,
+  stdout: ShellPipeWriter,
+  stderr: ShellPipeWriter,
+) -> ExecuteResult {
+  let result =
+    execute_pipeline_inner(pipeline.inner, state, stdin, stdout, stderr).await;
+  if pipeline.negated {
+    match result {
+      ExecuteResult::Exit(code, handles) => ExecuteResult::Exit(code, handles),
+      ExecuteResult::Continue(code, changes, handles) => {
+        let new_code = if code == 0 { 1 } else { 0 };
+        ExecuteResult::Continue(new_code, changes, handles)
+      }
+    }
+  } else {
+    result
+  }
+}
+
+async fn execute_pipeline_inner(
+  pipeline: PipelineInner,
+  state: ShellState,
+  stdin: ShellPipeReader,
+  stdout: ShellPipeWriter,
+  stderr: ShellPipeWriter,
+) -> ExecuteResult {
+  match pipeline {
+    PipelineInner::Command(command) => match command {
+      Command::Simple(command) => {
+        execute_simple_command(command, state, stdin, stdout, stderr).await
+      }
+      Command::Subshell(list) => {
+        execute_subshell(list, state, stdin, stdout, stderr).await
+      }
+    },
+    PipelineInner::PipeSequence(pipe_sequence) => {
+      execute_pipe_sequence(*pipe_sequence, state, stdin, stdout, stderr).await
+    }
+  }
+}
+
+async fn execute_pipe_sequence(
+  pipe_sequence: PipeSequence,
+  state: ShellState,
+  stdin: ShellPipeReader,
+  stdout: ShellPipeWriter,
+  stderr: ShellPipeWriter,
+) -> ExecuteResult {
+  let mut wait_tasks = vec![];
+  let mut last_output = Some(stdin);
+  let mut next_sequence = Some(Sequence::Pipeline(Pipeline {
+    negated: false,
+    inner: pipe_sequence.into(),
+  }));
+  while let Some(sequence) = next_sequence.take() {
+    let (output_reader, output_writer) = pipe();
+    let (stderr, sequence) = match sequence {
+      Sequence::Pipeline(Pipeline {
+        inner: PipelineInner::PipeSequence(pipe_sequence),
+        negated,
+      }) => {
+        assert!(!negated);
+        next_sequence = Some(pipe_sequence.next);
+        (
+          match pipe_sequence.op {
+            PipeSequenceOperator::Stdout => stderr.clone(),
+            PipeSequenceOperator::StdoutStderr => output_writer.clone(),
+          },
+          pipe_sequence.current,
+        )
+      }
+      _ => (stderr.clone(), sequence),
+    };
+    wait_tasks.push(execute_sequence(
+      sequence,
+      state.clone(),
+      last_output.take().unwrap(),
+      output_writer.clone(),
+      stderr.clone(),
+    ));
+    last_output = Some(output_reader);
+  }
+  let output_handle = tokio::task::spawn_blocking(|| {
+    last_output.unwrap().pipe_to_sender(stdout).unwrap();
+  });
+  let mut results = futures::future::join_all(wait_tasks).await;
+  output_handle.await.unwrap();
+  let last_result = results.pop().unwrap();
+  let all_handles = results.into_iter().flat_map(|r| r.into_handles());
+  match last_result {
+    ExecuteResult::Exit(code, mut handles) => {
+      handles.extend(all_handles);
+      ExecuteResult::Continue(code, Vec::new(), handles)
+    }
+    ExecuteResult::Continue(code, _, mut handles) => {
+      handles.extend(all_handles);
+      ExecuteResult::Continue(code, Vec::new(), handles)
+    }
+  }
+}
+
+async fn execute_subshell(
+  list: Box<SequentialList>,
+  state: ShellState,
+  stdin: ShellPipeReader,
+  stdout: ShellPipeWriter,
+  stderr: ShellPipeWriter,
+) -> ExecuteResult {
+  let result = execute_sequential_list(
+    *list,
+    state.clone(),
+    stdin,
+    stdout,
+    stderr,
+    // yield async commands to the parent
+    AsyncCommandBehavior::Yield,
+  )
+  .await;
+
+  // sub shells do not cause an exit
+  match result {
+    ExecuteResult::Exit(code, handles) => {
+      ExecuteResult::Continue(code, Vec::new(), handles)
+    }
+    ExecuteResult::Continue(_, _, _) => result,
+  }
+}
+
+async fn execute_simple_command(
+  command: SimpleCommand,
   state: ShellState,
   stdin: ShellPipeReader,
   mut stdout: ShellPipeWriter,
@@ -372,7 +428,11 @@ async fn execute_command(
       );
     }
 
-    let command_path = match resolve_command_path(&command_name, &state).await {
+    let command_path = match resolve_command_path(&command_name, &state, || {
+      Ok(std::env::current_dir()?)
+    })
+    .await
+    {
       Ok(command_path) => command_path,
       Err(err) => {
         stderr.write_line(&err.to_string()).unwrap();
@@ -420,11 +480,28 @@ async fn execute_command(
 async fn resolve_command_path(
   command_name: &str,
   state: &ShellState,
+  current_exe: impl FnOnce() -> Result<PathBuf>,
 ) -> Result<PathBuf> {
   if command_name.is_empty() {
     bail!("command name was empty");
   }
 
+  // Special handling to use the current executable for deno.
+  // This is to ensure deno tasks that use deno work in environments
+  // that don't have deno on the path and to ensure it use the current
+  // version of deno being executed rather than the one on the path,
+  // which has caused some confusion.
+  if command_name == "deno" {
+    if let Ok(exe_path) = current_exe() {
+      // this condition exists to make the tests pass because it's not
+      // using the deno as the current executable
+      let file_stem = exe_path.file_stem().map(|s| s.to_string_lossy());
+      if file_stem.map(|s| s.to_string()) == Some("deno".to_string()) {
+        return Ok(exe_path);
+      }
+    }
+  }
+
   // check for absolute
   if PathBuf::from(command_name).is_absolute() {
     return Ok(PathBuf::from(command_name));
@@ -634,3 +711,26 @@ async fn execute_with_stdout_as_text(
   let data = output_handle.await.unwrap();
   String::from_utf8_lossy(&data).to_string()
 }
+
+#[cfg(test)]
+mod test {
+  use super::*;
+
+  #[tokio::test]
+  async fn should_resolve_current_exe_path_for_deno() {
+    let state =
+      ShellState::new(Default::default(), &std::env::current_dir().unwrap());
+    let path =
+      resolve_command_path("deno", &state, || Ok(PathBuf::from("/bin/deno")))
+        .await
+        .unwrap();
+    assert_eq!(path, PathBuf::from("/bin/deno"));
+
+    let path = resolve_command_path("deno", &state, || {
+      Ok(PathBuf::from("/bin/deno.exe"))
+    })
+    .await
+    .unwrap();
+    assert_eq!(path, PathBuf::from("/bin/deno.exe"));
+  }
+}
